/**
 * @file TSPSolver.cpp
 * @brief TSP solver (neighborhood search)
 *
 */

#include "TSPSolver.h"
#include <iostream>

bool TSPSolver::solve ( const TSP& tsp , const TSPSolution& initSol , int tabulength , int maxIter , TSPSolution& bestSol )   /// TS: new param
{
  try
  {
    bool stop = false;
    int  iter = 0;

    ///Tabu Search
		tabuLength = tabulength;
		tabuList.reserve(tsp.n);
		initTabuList(tsp.n);
		///
		
		TSPSolution currSol(initSol);
    double bestValue, currValue;
    bestValue = currValue = evaluate(currSol,tsp);
    TSPMove move;
    while ( ! stop ) {
      ++iter;                                                                                             /// TS: iter not only for displaying
      if ( tsp.n < 20 ) currSol.print();
      std::cout << " (" << iter << ") value " << currValue << "\t(" << evaluate(currSol,tsp) << ")";      /// TS: iter
      double aspiration = bestValue-currValue;                                                            //**// TSAC: aspired improvement
		  double bestNeighValue = currValue + findBestNeighbor(tsp,currSol,iter,aspiration,move);             //**// TSAC: aspiration
      //if ( bestNeighValue < currValue ) {                                                               /// TS: replace stopping and moving criteria
      //  bestValue = currValue = bestNeighValue;                                                         ///
      //  currSol = swap(currSol,move);                                                                   ///
      //  stop = false;                                                                                   ///
      //} else {                                                                                          ///
      //  stop = true;                                                                                    ///
      //}                                                                                                 ///
      
      if ( bestNeighValue >= tsp.infinite ) {                                                             // TS: stop because all neighbours are tabu
        std::cout << "\tmove: NO legal neighbour" << std::endl;                                           //
        stop = true;                                                                                      //
        continue;                                                                                         //
      }                                                                                                   //
      
      std::cout << "\tmove: " << move.from << " , " << move.to;
 
      tabuList[currSol.sequence[move.from]] = iter;                                                       /// TS: update tabu list
      tabuList[currSol.sequence[move.to]]   = iter;                                                       ///
      currSol = swap(currSol,move);                                                                       /// TS: always the best move
      currValue = bestNeighValue;                                                                         /// 
      if ( currValue < bestValue -0.01 ) {                                                                /// TS: update incumbent (not always)
        bestValue = currValue;                                                                            ///
        bestSol = currSol;                                                                                ///
        std::cout << "\t***";                                                                             ///
		  }                                                                                                   ///
		  
		  if ( iter > maxIter ) {                                                                             /// TS: new stopping criteria
  			stop = true;                                                                                      ///
		  }                                                                                                   ///
      std::cout << std::endl;  
    }
    //bestSol = currSol;                                                                                  /// TS: not always improves
  }
  catch(std::exception& e)
  {
    std::cout << ">>>EXCEPTION: " << e.what() << std::endl;
    return false;
  }
  return true;
}

TSPSolution& TSPSolver::swap ( TSPSolution& tspSol , const TSPMove& move ) 
{
  TSPSolution tmpSol(tspSol);
  for ( int i = move.from ; i <= move.to ; ++i ) {
    tspSol.sequence[i] = tmpSol.sequence[move.to-(i-move.from)];
  }
  return tspSol;
}


double TSPSolver::findBestNeighbor ( const TSP& tsp , const TSPSolution& currSol , int currIter , double aspiration , TSPMove& move )     //**// TSAC: use aspiration
{
  double bestDecrement = tsp.infinite;

  for ( uint a = 1 ; a < currSol.sequence.size() - 2 ; a++ ) {
    int h = currSol.sequence[a-1];
    int i = currSol.sequence[a];
    for ( uint b = a + 1 ; b < currSol.sequence.size() - 1 ; b++ ) {
      int j = currSol.sequence[b];
      int l = currSol.sequence[b+1];
      //**// TASC: after... if ( (currIter - tabuList[i] <= tabuLength) && (currIter - tabuList[j] <= tabuLength) ) continue;
	    double neighDecrement = - tsp.cost[h][i] - tsp.cost[j][l]
                              + tsp.cost[h][j] + tsp.cost[i][l] ;
      if ( (currIter - tabuList[i] <= tabuLength) && (currIter - tabuList[j] <= tabuLength) && !(neighDecrement < aspiration-0.01) ) {
		    continue;             //**// TSAC: check if tabu and not aspiration criteria
		  } 
      if ( neighDecrement < bestDecrement ) {
        bestDecrement = neighDecrement;
        move.from = a;
        move.to = b;
      }
      //if ( bestDecrement < 0 ) return bestDecrement; //*****// First Improvement
    }
  }
  if ( bestDecrement >= tsp.infinite ) std::cout << "\n AARRGH!!! " << std::endl;
  return bestDecrement;
}
